/*
  metautil.c: utilities for baysian fitting the patch occupancy model of
   metapopulation dynamics
   C.J.F. ter Braak, CBW, Wageningen
   c.j.f.terbraak@plant.wag-ur.nl
   June 27, 2000
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <float.h>
#include <limits.h>

/* numerical recipes: */



#include <nrutil.h>

#include <nr.h>

#include "fmetapop.h"


extern FILE *out, *spc, *dat, *mon;


int sum_row_imatrix(int r, int c, int **imat, int missing, int sum[])
 /*  returns the matrix total and, in r-vector sum, the total of rows in rxc imatrix imat 
  for the non-missing data
	Missing data are indicated by missing
	imat is one-based
 
 */
{
	int i, j, tot=0, sumi;
  for (i =1;i<=r ;i++ ) {
	  sumi = 0;
   	for (j=1;j<=c ;j++ ){
 		  if (imat[i][j]!=missing)  sumi+=imat[i][j];
		}
		tot+= sumi;
		sum[i]= sumi; 	
  }
  
	return tot;

}
void swap_pd(double **p0, double **p1)
/* swap pointers to double-arrays
   Call as: swap_dp(&p0,&p1)
	 Beware: if swapping rows of an NR-matrix, make sure the first row points to
	 the original one on calling free_dmatrix, e.g. by making the lower row index (rl) 
	 one lower than needed in the program
	 
	  */
{ 
	double  *p;
	p  = *p1;
	*p1 = *p0;
	*p0 = p;
}

void swap_pi(int **p0, int **p1)
/* swap pointers to int-arrays 
   Call as: swap_di(&p0,&p1) 
	 Beware: if swapping rows of an NR-matrix, make sure the first row points to
	 the original one on calling free_imatrix, e.g. by making the lower row index (rl) 
	 one lower than needed in the program
	 
	  */
{ 
	int  *p;
	p  = *p1;
	*p1 = *p0;
	*p0 = p;
}






void transpose_dmatrix(double **a, double **at, int nr, int nc )
/* transposes  a[1...nr][1...nc] in to at[1...nc][1...nr] */
{
 int i,j;
 for (i=1;i<=nr ;i++ ) {
 	for (j=1; j<=nc ;j++ ) { 
		at[j][i] = a[i][j];
	}
 }
}

void  multiply_dmat_const( double **a, double b, double **ba, int rows_a, int cols_a)

/* for one-based  double matrices: multiplication of a matrix a by a constant b  */

{ 
  int r,c;

  for (r=1; r<=rows_a; r++)
    for (c=1; c<=cols_a; c++)
      ba[r][c]= b*a[r][c];

} /* multiply_dmat_const() */



void multiply_dmatrices ( double **a, double **b, double **ab,
                        int rows_a, int cols_a, int cols_b )

/* for one-based  double matrices rows_b==cols_a,  rows_ab==rows_a,  cols_ab==cols_b  */

{ 
  int i,r,c;

  for (r=1; r<=rows_a; r++)
    for (c=1; c<=cols_b; c++)
      for (ab[r][c]=0.0,i=1; i<=cols_a; i++)  ab[r][c]+= a[r][i]*b[i][c];

} /* multiplydmatrices() */




void mult_dmatvec ( double **m, double *v, double *mv, int rows, int cols )
/*  for one-based a matrix m and vector v:
    mv = m v
		*/
{ 
  int r,c;

  for (r=1; r<=rows; r++)
    for (mv[r]=0.0, c=1; c<=cols; c++)  mv[r]+= m[r][c]*v[c];

} /* multdmatvec() */

void mult_dlowermatvec ( double **m, double *v, double *mv, int rows)
/*  for one-based a lower triangular matrix m and vector v:
    mv = m v
		*/
{ 
  int r,c;

  for (r=1; r<=rows; r++)
    for (mv[r]=0.0, c=1; c<=r; c++)  mv[r]+= m[r][c]*v[c];

} /* multdmatvec() */





void copy_dmatrix ( double **source, double **dest, int rows, int cols )
/* copy for one-based matrices */
{ 
  int r,c;

  for (r=1; r<=rows; r++)
    for (c=1; c<=cols; c++)  dest[r][c]= source[r][c];

} /* copydmatrix() */

void copy_dvector ( double *source, double *dest, int n )
/* copy for one-based vectors */
{ 
  int r;

  for (r=1; r<=n; r++) dest[r]= source[r];

} /* copydvector() */





int  cholesky(double** s, double** l, int n)
/* Given a positive-definiet symmetric matrix s[1...n][1...n] this routine constructs
   its Choleski decompositon s = l l' 
	 From the input s only its lower triangle is used in the calculations, diagonal inclusive
	 The choleski factor l is stored in lower triangular form, diagonal inclusive
	 The upper triangle of l is set to 0, as required. 
	 If s is not positive definite an error and 0 is returned
 	*/
{
   int i, j, k;
   double x;
   for (j=1; j<=n; j++) for (i=j; i<=n; i++){
      x = s[i][j]; 
      for (k=1; k<=j-1; k++) x -= l[j][k] * l[i][k];
      if (i==j){
         if (x<=0){
				      message("cholesky: matrix not strictly positive");
							return 0;
				 }
				 l[i][i] = sqrt(x);
      }
      else { 				 
					l[i][j] = x / l[j][j];
					l[j][i] = 0;
			}
   }
	 return 1;
}



void read_lower_matrix(FILE *in, double **cov, int p)
/* reads a lower matrix cov from file in and sets the upper part */
{
 int r,c;
 for (r=1; r<=p ;r++ ) {
 	 for (c=1;c<=r ;c++ ) {
    	fscanf(in,  " %lf " ,&cov[r][c]);
		  cov[c][r] =cov[r][c];
	}
 }
 # if DEBUG
 		fprintfdmatrix(out, cov,  1,p, 1,p, "%.2lf ");
 # endif
}

void set_cov_diag(double **cov, int p, double *diag_values)
/* sets the square matrix cov to diag(diag_values^2) */
{
	int r,c;
	for (r=1; r<=p ;r++ ) {for (c=1;c<=p ;c++ ) cov[r][c] = 0.0;}
	for (r=1; r<=p; r++ ) cov[r][r] = diag_values[r]; 
}

int *create_vindex(int n, double *x, int *ncrit, double crit)
/*  creates a vector-index of length ncrit
		 indicating elements of a double vector x that are equal to the value  crit
	   ending list with a zero 
		 The values returned assume that x is x[1...n] */
{
   int i, ii=0, nelem=0;
	 int *list;
	 for (i=1;i<=n ;i++ ) {
	 	if (fabs(x[i] - crit )> FLT_MIN) {
			nelem++;
		}
	 }
	 list = ivector(1,nelem+1);
	 for (i=1;i<=n ;i++ ) {
	 	if (fabs(x[i] - crit )> FLT_MIN) {
			ii++;
			list[ii] = i;
		}
	 }
	 *ncrit = nelem;
	 return list;



}
 

int **create_missinglist(int r, int c, int **idata, int *missing)
 /*  returns number of missings and creates list of missings (r,c) in the data (idata)
outputted as
 rc_missinglist 	= list of length nmissing+1 of missing values (MISSING (-1)) in idata:
 														 col1,col2 =(r,c)  ending with (INT_MAX,INT_MAX)
														  
rc_missinglist is allocated in this function
*/
{
	int i,j, t, nmissing= 0;
	int **rc_missinglist;
   for (t =1;t<=r ;t++ ) {
    	for (i=1;i<=c ;i++ ) if (idata[t][i]==MISSING) nmissing++ ;
   }
  rc_missinglist = imatrix(1, nmissing+1,1,2);
  if (nmissing){
  j=0;
  for (t =1;t<=r ;t++ ) {
   	for (i=1;i<=c ;i++ ){
 		  if (idata[t][i]==MISSING) {
        j++;
 		    rc_missinglist[j][1] =t ;
 		    rc_missinglist[j][2] =i ;
 		} 	
    }
  }
  }
  rc_missinglist[nmissing+1][1] = INT_MAX;
  rc_missinglist[nmissing+1][2] = INT_MAX;

	/*fprintfimatrix(out, rc_missinglist,1,nmissing+1,1,2, "%3d");*/
	*missing = nmissing;
	return rc_missinglist;

}




int read_file(int npatch, int nyear, double *xcoor, double *ycoor,\
 		double *A, int **idata, int nxpar, double  **X)
{
	int i,j,t;
 /*  order of data, for each patch:
	 x_coordinate, y_coordinate, area, X_variables for E (if any), X_variables for C (if any),
	      1/0/-1 incidence data indicating occupancy/empty/missing for each year (1... nyear)
 
*/

 /* reading input data */
   for ( i=1; i<=npatch; i++) {
     fscanf(dat, "%lf %lf %lf", &(xcoor[i]), &(ycoor[i]), &(A[i]));
       for (j=1; j<=nxpar ; j++ ) fscanf(dat, "%lf", &(X[i][j]));
       for (t=1; t<=nyear ; t++ ) fscanf(dat, "%d", &(idata[t][i]));
      
   }
	#if DEBUG
	/* output inputdata */
	   for ( i=1; i<=npatch; i++) {
	     fprintf(out, "\n%8.2lf %8.2lf %8.2lf", xcoor[i], ycoor[i], A[i]);
	     for (t=1; t<=nyear ; t++ ) {
	       fprintf(out, "%3d", idata[t][i]);
	     }
	   }
	     fprintf(out, "\n");
	#endif

	 return 1;
}


float smpos(float x, float scale)
{
   float z, a;
   a = log(2);
   z = a * x / scale;
   if (fabs(z)<=30) return scale * log(1+exp(a*x/scale)) / a;
   else if (z<-30) return 0;
   else return x;
}

int any(int n1, int n2, long* ps)
{
   float t;
   t = (n2-n1+1) *urand(ps);
   if (t==0) return n1;
   else return n1-1+ceil(t);
}


void error(char why[])
{
   fprintf(out,"\n***Run-time error***...\n");
   fprintf(out,"%s\n", why);
   fprintf(out,"...bye-bye...\n");
   exit(1);
}

void message(char why[])
{
   fprintf(out,"\n***Run-time message***: ");
   fprintf(out,"%s\n", why);
}

void meta_message(FILE *out, char why[])
{
   fprintf(out,"***Run-time message***: ");
   fprintf(out,"%s\n", why);
}


void print_time(FILE *out)
      {
        struct tm *ptr;
        time_t lt;
        lt = time(NULL);
        ptr = localtime(&lt);
        fprintf(out, "\nTime: ");
        fprintf(out, asctime(ptr));
      }


void readf(FILE *in, FILE *out, char* identifier, int npar, double* x, double minx, double maxx, \
                                       double* y, double miny, double maxy)
/* read two doubles (x and y) from file in between given boundaries */
{
 char error_text[100],iden[40];
 if (npar ==1) {
   if ((fscanf(in, "%s %lf", iden, x))!=2) {
  	  sprintf(error_text,  "\nError reading %10s ",identifier);
  		error(error_text);
     }
   if (strcmp(identifier,iden)) {
  	  sprintf(error_text,  "\nError: Expecting identifier %10s, while reading %10s ",identifier, iden);
  		error(error_text);
   }
   if (*x < minx) {
  	  sprintf(error_text,  "\nError: %10s less than the minimum value %.2lf ",identifier, minx);
  		error(error_text);
   }
   if (*x  > maxx) {
  	  sprintf(error_text,  "\nError: %10s greater than the maximum value %.2lf ",identifier, minx);
  		error(error_text);
   }
   fprintf(out,"\nParameter %s = %.2lf", identifier, *x);
 }
  else if (npar == 2) {
   if ((fscanf(in, "%s %lf %lf", iden, x, y))!=3) {
  	  sprintf(error_text,  "\nError reading %10s ",identifier);
  		error(error_text);
   }  
   if (strcmp(identifier,iden)) {
  	  sprintf(error_text,  "\nError: Expecting identifier %10s, while reading %10s ",identifier, iden);
  		error(error_text);
   }
   if (*x < minx) {
  	  sprintf(error_text,  "\nError: %10s less than the minimum value %.2lf ",identifier, minx);
  		error(error_text);
   }
   if (*x  > maxx) {
  	  sprintf(error_text,  "\nError: %10s greater than the maximum value %.2lf ",identifier, minx);
  		error(error_text);
   }
   if (*y < miny) {
  	  sprintf(error_text,  "\nError: %10s less than the minimum value %.2lf ",identifier, miny);
  		error(error_text);
   }
   if (*y  > maxy) {
  	  sprintf(error_text,  "\nError: %10s greater than the maximum value %.2lf ",identifier, miny);
  		error(error_text);
   }
   fprintf(out,"\nParameter %s = %.2lf with second value %.2lf", identifier, *x, *y);  }
  else {
		 error("Fault in npar of readf");
	}

}
void readl(FILE *in, FILE *out, char* identifier, long npar, long* x, long minx, long maxx, \
                                       long* y, long miny, long maxy)
/* read two longs (x and y) from file in between given boundaries */
{
 char error_text[100],iden[40];
 if (npar ==1) {
   if ((fscanf(in, "%s %ld", iden, x))!=2) {
  	  sprintf(error_text,  "\nError reading %10s ",identifier);
  		error(error_text);
     }
   if (strcmp(identifier,iden)) {
  	  sprintf(error_text,  "\nError: Expecting identifier %10s, while reading %10s ",identifier, iden);
  		error(error_text);
   }
   if (*x < minx) {
  	  sprintf(error_text,  "\nError: %10s less than the minimum value %ld ",identifier, minx);
  		error(error_text);
   }
   if (*x  > maxx) {
  	  sprintf(error_text,  "\nError: %10s greater than the maximum value %ld ",identifier, minx);
  		error(error_text);
   }
   fprintf(out,"\nParameter %s = %ld", identifier, *x);
 }
  else if (npar == 2) {
   if ((fscanf(in, "%s %ld %ld", iden, x, y))!=3) {
  	  sprintf(error_text,  "\nError reading %10s ",identifier);
  		error(error_text);
   }  
   if (strcmp(identifier,iden)) {
  	  sprintf(error_text,  "\nError: Expecting identifier %10s, while reading %10s ",identifier, iden);
  		error(error_text);
   }
   if (*x < minx) {
  	  sprintf(error_text,  "\nError: %10s less than the minimum value %ld ",identifier, minx);
  		error(error_text);
   }
   if (*x  > maxx) {
  	  sprintf(error_text,  "\nError: %10s greater than the maximum value %ld ",identifier, minx);
  		error(error_text);
   }
   if (*y < miny) {
  	  sprintf(error_text,  "\nError: %10s less than the minimum value %ld ",identifier, miny);
  		error(error_text);
   }
   if (*y  > maxy) {
  	  sprintf(error_text,  "\nError: %10s greater than the maximum value %ld ",identifier, miny);
  		error(error_text);
   }
   fprintf(out,"\nParameter %s = %ld with second value %ld", identifier, *x, *y);  }
  else {
		 error("Fault in npar of readi");
	}

}
void readi(FILE *in, FILE *out, char* identifier, int npar, int* x, int minx, int maxx, \
                                       int* y, int miny, int maxy)
/* read two ints (x and y) from file in between given boundaries */
{
 char error_text[100],iden[40];
 if (npar ==1) {
   if ((fscanf(in, "%s %d", iden, x))!=2) {
  	  sprintf(error_text,  "\nError reading %10s ",identifier);
  		error(error_text);
     }
   if (strcmp(identifier,iden)) {
  	  sprintf(error_text,  "\nError: Expecting identifier %10s, while reading %10s ",identifier, iden);
  		error(error_text);
   }
   if (*x < minx) {
  	  sprintf(error_text,  "\nError: %10s less than the minimum value %d ",identifier, minx);
  		error(error_text);
   }
   if (*x  > maxx) {
  	  sprintf(error_text,  "\nError: %10s greater than the maximum value %d ",identifier, minx);
  		error(error_text);
   }
   fprintf(out,"\nParameter %s = %d", identifier, *x);
 }
  else if (npar == 2) {
   if ((fscanf(in, "%s %d %d", iden, x, y))!=3) {
  	  sprintf(error_text,  "\nError reading %10s ",identifier);
  		error(error_text);
   }  
   if (strcmp(identifier,iden)) {
  	  sprintf(error_text,  "\nError: Expecting identifier %10s, while reading %10s ",identifier, iden);
  		error(error_text);
   }
   if (*x < minx) {
  	  sprintf(error_text,  "\nError: %10s less than the minimum value %d ",identifier, minx);
  		error(error_text);
   }
   if (*x  > maxx) {
  	  sprintf(error_text,  "\nError: %10s greater than the maximum value %d ",identifier, minx);
  		error(error_text);
   }
   if (*y < miny) {
  	  sprintf(error_text,  "\nError: %10s less than the minimum value %d ",identifier, miny);
  		error(error_text);
   }
   if (*y  > maxy) {
  	  sprintf(error_text,  "\nError: %10s greater than the maximum value %d ",identifier, miny);
  		error(error_text);
   }
   fprintf(out,"\nParameter %s = %d with second value %d", identifier, *x, *y);  }
  else {
		 error("Fault in npar of readi");
	}
}


